{
* qsys (unit)
* Copyright (c) 2004-2005 Andreas K. Foerster <akfquiz@akfoerster.de>
*
* Environment: FreePascal or GNU-Pascal
*
* This file is part of AKFQuiz
*
* AKFQuiz is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* AKFQuiz is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
*}

{ system specific stuff for AKFQuiz }

{$X+}
{$I-}

{$IfDef FPC}
  {$Mode Delphi}
  {$Smartlink on}
  {$LongStrings on}
{$EndIf}


{ compatiblity definition }
{$IfDef _WIN32} {$Define Win32} {$EndIf}


unit qsys;

interface

{$IfDef __GPC__}

  import GPC (GetEnv => GetEnvironmentVariable,
              FSearch => FileSearch,
	      FExpand => ExpandFilename,
	      DirFromPath => ExtractFilePath,
              DirSeparator => DirectorySeparator,
              LineBreak => LineEnding);
  {$IfDef Win32}
         Winprocs (GetEnvironmentVariable => WGetEnvironmentVariable); 
	 WinTypes;
  {$EndIf}

{$Else}

    uses
    {$IfDef Win32}
       Windows,
    {$EndIf} { Win32 }
    SysUtils;

{$EndIf} { __GPC__ }


{$IfDef FPC}
    type mystring = ansistring; { Delphi dialect (unlimited length) }
{$Else}
    type mystring = string(2048); { Extended Pascal }
{$EndIf}

{ strings are implemented differently in Pascal dialects. }
type string75  = string[75];
type string80  = string[80];
type string255 = string[255];

type shortstring = string255; { needed by GPC }

type Unicode = Cardinal; { Large integer value - larger than 3 Bytes }

type Tshowentry = procedure(const dir, f : string);
type Tconverter = function(const s: string): mystring;

{ Displays for which converters are there }
type DisplayType = (ISOdisplay, OEMdisplay, UTF8display);

type 
  PquizfileList = ^TquizfileList;
  TquizfileList = record
                  filename : mystring;
                  title    : string255;
                  next     : PquizfileList
                  end;

{ only used on very few systems - on most it's autodetected }
{$IfDef UNIX}
  const defaultPrefix = '/usr/local';
{$Else}
  const defaultPrefix = '';
{$EndIf}

const quizext  = '.akfquiz';
const quizext2 = '.aqz';

const platform =
{$IfDef __GPC__} 'GPC'; {$EndIf}
{$IfDef FPC}     'FPC '+{$I %FPCTARGETOS%}+'/'+{$I %FPCTARGETCPU%};
{$EndIf}

{ what platforms have charset IBM850 as default? }
{$IfDef DPMI}       {$Define IBM850} {$EndIf}
{$IfDef __OS_DOS__} {$Define IBM850} {$EndIf}
{$IfDef MSDOS}      {$Define IBM850} {$EndIf}

{ IBM850 is supported by more webbrowsers than IBM437,
  but in most every case it's okay to pretend it's 850 }

{$IfDef IBM850}
  const sys_charset = 'IBM850'; { see comment above }
{$Else}
  {$IfDef Win32}
    const sys_charset = 'WINDOWS-1252';
  {$Else}
    const sys_charset = 'ISO-8859-1';
  {$EndIf}
{$EndIf}

const TAB = chr(9);
const nl = LineEnding;

{ HTML 4.01 is the last HTML definition
  Transitional is needed, because of the "target" attribute }
const HTML4DTD = 'http://www.w3.org/TR/REC-html4/loose.dtd';
const HTMLDoctype =
    '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"'+nl+
    '"'+HTML4DTD+'">';

{ Attetion:
  XHTML conflicts with the JavaScript implementation! }
const XHTML1DTD = 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd';
const XHTMLDoctype =
    '<?xml version="1.0" encoding="iso-8859-1"?>'+nl+
    '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"'+nl+
    '"'+XHTML1DTD+'">';

const xmlns='http://www.w3.org/1999/xhtml';

var 
  IntroSignal,    { for Titlescreen }
  QuestionSignal, { for filechooser }
  RightSignal,
  FalseSignal,
  NeutralSignal,
  InfoSignal,  { for result screen }
  ErrorSignal: procedure;

var quizfileList, quizfileListEnd : PquizfileList;

function makeUpcase(x: string): mystring;
function stripWhitespace(x: string): mystring;
function Min(a, b: LongInt): LongInt;
function Max(a, b: LongInt): LongInt;
{ remove to last dot }
function stripext(const s: string): mystring;

function basename(const s: string): mystring;
function getnextdir(var rest: mystring): mystring;
function getquizpath: mystring;
function getquizdir: mystring; { first dir from QUIZPATH }
function useDirSeparator(const s: string): mystring;
function getQuizTitle(const x: string): mystring;

{ searches quizfile: }
function getquizfile(var s: mystring): boolean; 
function quizfileExists(const s: string): boolean;
function gethtmlname(s: mystring): mystring;
function ListEntries(const dir, ext: string; showentry: Tshowentry): boolean;

procedure freeQuizFileList;
procedure addQuizFileList(const dir, ext: string);
procedure fillQuizfileList; { uses quizpath! }
function searchQuizfileList(searchNr: integer): mystring;

procedure nobreak;
procedure setLFNsupport;

function checkISO(const s: mystring): boolean;
function checkUTF8(const s: mystring): boolean;
function checkOEM(const s: mystring): boolean;
function checkASCII(const s: mystring): boolean;
function checkCP1252(const s: mystring): boolean;

{ check what the display supports }
function checkOEMdisplay: boolean;
function checkUTF8display: boolean;
function getSystemLanguage: mystring;

{ procedures for signal variables }
procedure NoSignal;
procedure SystemBeep;

{ assign all signals: }
procedure DisableSignals;
procedure useBeepSignals; { use system beeps (#7) }

procedure setUnknownChar(c: char);

{ for Tconverter }
{ OEM means IBM850, but it can also be used for IBM437 }

function Latin1toUTF8(const s: string): mystring;
function UTF8toLatin1(const s: string): mystring;
function UTF8toLatin9(const s: string): mystring;
function Latin1toOEM(const s: string): mystring;
function OEMtoLatin1(const s: string): mystring;
function OEMtoUTF8(const s: string): mystring;
function UTF8toOEM(const s: string): mystring;
function CP1252toLatin9(const s: string): mystring;
{ HTML numeric entities: }
function Latin1toHTML(const s: string): mystring;
function UTF8toHTML(const s: string): mystring;
function forceASCII(const s: string): mystring;
function noconversion(const s: string): mystring;

{ length of real characters in a UTF8 string }
function UTF8Length(const s: string): LongInt;

function ShowTime(sec: LongInt): mystring;
function GetSecs: LongInt;

{$IfDef __GPC__}
  function IntToStr(i: LongInt): mystring;
{$EndIf}



implementation

{ I'm still using FPC 1.0.10 under DOS due to problems with newer versions }
{$IfDef FPC}{$IfDef VER1_0}
  {$INFO 1_0}
  uses DOS; { for GetEnv }
{$EndIf}{$EndIf}

const 
  CP850Unicode: array[#$80..#$FF] of Unicode = (
     $00c7, $00fc, $00e9, $00e2, $00e4, $00e0, $00e5, $00e7,
     $00ea, $00eb, $00e8, $00ef, $00ee, $00ec, $00c4, $00c5,
     $00c9, $00e6, $00c6, $00f4, $00f6, $00f2, $00fb, $00f9,
     $00ff, $00d6, $00dc, $00f8, $00a3, $00d8, $00d7, $0192,
     $00e1, $00ed, $00f3, $00fa, $00f1, $00d1, $00aa, $00ba,
     $00bf, $00ae, $00ac, $00bd, $00bc, $00a1, $00ab, $00bb,
     $2591, $2592, $2593, $2502, $2524, $00c1, $00c2, $00c0,
     $00a9, $2563, $2551, $2557, $255d, $00a2, $00a5, $2510,
     $2514, $2534, $252c, $251c, $2500, $253c, $00e3, $00c3,
     $255a, $2554, $2569, $2566, $2560, $2550, $256c, $00a4,
     $00f0, $00d0, $00ca, $00cb, $00c8, $0131, $00cd, $00ce,
     $00cf, $2518, $250c, $2588, $2584, $00a6, $00cc, $2580,
     $00d3, $00df, $00d4, $00d2, $00f5, $00d5, $00b5, $00fe,
     $00de, $00da, $00db, $00d9, $00fd, $00dd, $00af, $00b4,
     $00ad, $00b1, $2017, $00be, $00b6, $00a7, $00f7, $00b8,
     $00b0, $00a8, $00b7, $00b9, $00b3, $00b2, $25a0, $00a0);

var QUIZPATH : mystring;
var unknownChar : char; { replacement for untranslatable characters }

{$IfDef FPC}{$IfDef VER1_0}
  function GetEnvironmentVariable(s: string): mystring;
  begin
  GetEnvironmentVariable := GetEnv(s)
  end;
{$EndIf}{$EndIf}

function Min(a, b: LongInt): LongInt;
begin
if a<b then Min := a else Min := b
end;

function Max(a, b: LongInt): LongInt;
begin
if a>b then Max := a else Max := b
end;

{ removes leading and trailing spaces and tabs }
function stripWhitespace(x: string): mystring;
var len, f, t: integer;
begin
len := length(x);
f := 1;
t := len;
while (f<len) and ((x[f]=' ') or (x[f]=TAB)) do f := succ(f);
while (t>1) and ((x[t]=' ') or (x[t]=TAB) or (x[t]=chr(13)))
  do t := pred(t);
if f<=t 
  then stripWhitespace := copy(x, f, t)
  else stripWhitespace := ''
end;

function makeUpcase(x: string): mystring;
begin
{$IfDef __GPC__}
  makeUpcase := UpcaseStr(x)
{$Else}
  makeUpcase := Uppercase(x)
{$EndIf}
end;

procedure NoSignal;
begin end;

procedure SystemBeep;
begin Write(#7) end;

procedure DisableSignals;
begin
IntroSignal    := NoSignal;
QuestionSignal := NoSignal;
RightSignal    := NoSignal;
FalseSignal    := NoSignal;
NeutralSignal  := NoSignal;
InfoSignal     := NoSignal;
ErrorSignal    := NoSignal
end;

procedure useBeepSignals;
begin
IntroSignal    := NoSignal;
QuestionSignal := NoSignal;
RightSignal    := NoSignal;
FalseSignal    := SystemBeep;
NeutralSignal  := NoSignal;
InfoSignal     := SystemBeep;
ErrorSignal    := SystemBeep
end;

function getSystemLanguage: mystring;
var l: mystring;
begin
l := GetEnvironmentVariable('LANG');

{$IfDef Win32}
  if l='' then
    case (GetUserDefaultLangID and $03FF) of
      { GPC doesn't have the LANG constants yet }
      $36 : l := 'af';
      $04 : l := 'zh';
      $05 : l := 'cs';
      $06 : l := 'da';
      $13 : l := 'nl';
      $09 : l := 'en';
      $0B : l := 'fi';
      $0C : l := 'fr';
      $07 : l := 'de';
      $08 : l := 'el';
      $0D : l := 'he';
      $10 : l := 'it';
      $11 : l := 'jp';
      $14 : l := 'no';
      $15 : l := 'pl';
      $16 : l := 'pt';
      $19 : l := 'ru';
      $0A : l := 'es';
      $1D : l := 'sv';
      $1F : l := 'tr';
      { ... to be continued @@ }
      end;
{$EndIf}

getSystemLanguage := l
end;

function useDirSeparator(const s: string): mystring;
begin
{$IfDef __GPC__}
  useDirSeparator := ForceAddDirSeparator(s)
{$Else}
  useDirSeparator := IncludeTrailingPathDelimiter(s)
{$EndIf}
end;

{$IfDef NoRelocation}

  function getprefix: mystring;
  begin
  getprefix := defaultPrefix
  end;

{$Else} { Relocation }

  function searchExecutable(const s: mystring): mystring;
  begin
  searchExecutable := FileSearch(s, GetEnvironmentVariable('PATH'))
  end;

  function getprefix: mystring;
  var s: mystring;
  begin
  s := ParamStr(0);
  if s='' 
    then getprefix := defaultPrefix
    else begin
         if pos(DirectorySeparator, s)=0 { no directory given }
           then s := searchExecutable(s);
         getprefix := ExpandFileName(ExtractFilePath(s)+'..')
         end
  end;
 
{$EndIf} { NoRelocation }


{ search for Parameter -d }
function getparamdir: mystring;
var i, count : integer;
begin
{ search for Parameter -d }
count := ParamCount;
i:=1;
while (i<count) 
      and (ParamStr(i)<>'-d') 
      and (ParamStr(i)<>'-D') 
        do inc(i);
if (i<count) and ((ParamStr(i)='-d') or (ParamStr(i)='-D')) 
  then getparamdir := ParamStr(i+1)
  else getparamdir := ''
end;

function getnextdir(var rest: mystring): mystring;
var p: integer;
begin
p := pos(PathSeparator, rest);
if p<>0 
  then begin
       getnextdir := copy(rest, 1, p-1);
       delete(rest, 1, p)
       end
  else begin
       getnextdir := rest;
       rest := ''
       end
end;

function getquizpath: mystring;
begin
getquizpath := QUIZPATH
end;

function getquizdir: mystring;
var d: mystring;
begin
{ must be in a separate variable for getnextdir changes the content }
d := QUIZPATH;

{ only one directory! }
getquizdir := getnextdir(d)
end;

function getquizfile(var s: mystring): boolean;
var e, path: mystring;
begin
{ only search, when s hasn't a path yet }
if pos(DirectorySeparator, s)<>0 
  then getquizfile := FileExists(s)
  else begin
       path := QUIZPATH;
       e := FileSearch(s, path);
       if e='' then
          e := FileSearch(s+quizext, path);
       if e='' then
          e := FileSearch(s+quizext2, path);
       s := e;
       getquizfile := (e<>'')
       end
end;

function quizfileExists(const s: string): boolean;
begin
quizfileExists := FileExists(s) and not DirectoryExists(s)
end;

{ remove to last dot }
function stripext(const s: string): mystring;
var i: integer;
begin
i:=length(s);
while (i>1) and (s[i]<>'.') do i := i - 1;
dec(i);
if i>1 then stripext := copy(s, 1, i)
       else stripext := s
end;

procedure setUnknownChar(c: char);
begin
unknownChar := c
end;

function ISO8859ControlChar(u: Unicode): boolean;
begin
ISO8859ControlChar := ((u and $FFFFFF7F) < $20)
end;

function ASCIIControlChar(u: Unicode): boolean;
begin
ASCIIControlChar := (u < $20)
end;

{ functions for Tconvert }

function noconversion(const s: string): mystring;
begin
noconversion := s
end;

function forceASCII(const s: string): mystring;
var 
  i: integer;
  e: mystring;
begin
e := s;
for i := 1 to Length(e) do 
  if (e[i]<#32) or (e[i]>#127) then e[i] := unknownChar;
forceASCII := e
end;

function EncodeUTF8(u: Unicode): string80;
begin
case u of
  $000000..$00007F : EncodeUTF8 := chr(u);
  $000080..$0007FF : EncodeUTF8 := chr($C0 or (u shr 6)) +
                                   chr($80 or (u and $3F));
  $000800..$00FFFF : EncodeUTF8 := chr($E0 or (u shr (2*6))) +
                                   chr($80 or ((u shr 6) and $3F)) +
                                   chr($80 or (u and $3F));
  $010000..$1FFFFF : EncodeUTF8 := chr($F0 or (u shr (3*6))) +
                                   chr($80 or ((u shr (2*6)) and $3F)) +
                                   chr($80 or ((u shr 6) and $3F)) +
                                   chr($80 or (u and $3F));
  otherwise EncodeUTF8 := unknownChar
  end
end;

{ gets the Unicode value of the first char in the UTF-8 string
  and removes the first char from the string 
  RFC 3629, ISO 10646 }
function getUTF8char(var s: mystring): Unicode;
var 
  i : integer;
  u : Unicode;
begin
u := ord(unknownChar); 

{ attention: do not use this decoder in security critical areas
  it also decodes invalid UTF-8 encodings }
  
if s<>'' then 
  begin
  i := ord(s[1]); { save it, since "s" is changed later }
  case i of
     $00..$7F : begin { 1 byte encoding }
                u := ord(s[1]);
                Delete(s, 1, 1)
                end;
     $C2..$DF : begin { 2 byte encoding }
                u := (ord(s[1]) and $1F) shl 6;
                u := u or (ord(s[2]) and $3F);
                Delete(s, 1, 2)
                end;
     $E0..$EF : begin { 3 byte encoding }
                u := (ord(s[1]) and $0F) shl (2*6);
                u := u or ((ord(s[2]) and $3F) shl 6);
                u := u or (ord(s[3]) and $3F);
                Delete(s, 1, 3)
                end;
     $F0..$F7 : begin { 4 byte encoding }
                u := (ord(s[1]) and $07) shl (3*6);
                u := u or ((ord(s[2]) and $3F) shl (2*6));
                u := u or ((ord(s[3]) and $3F) shl 6);
                u := u or (ord(s[4]) and $3F);
                Delete(s, 1, 4)
                end;
     otherwise Delete(s, 1, 1) { remove unknown char anyway }
     end { case }
  end; { if s<>'' }

getUTF8char := u
end;

function Latin1toUTF8(const s: string): mystring;
var
  i : integer;
  e : mystring;
begin
{ Latin1 is a subset to unicode, so ord() reveals the unicode number }
e := '';
for i := 1 to length(s) do
  e := e + EncodeUTF8(ord(s[i]));
Latin1toUTF8 := e
end;

function UTF8toHTML(const s: string): mystring;
var 
  f, e : mystring;
  u: Unicode;
begin
f := s;
e := '';

while f <>'' do 
  begin
  u := getUTF8char(f);
  if u<=127 
    then e := e + chr(u) { ASCII }
    else e := e + '&#' + IntToStr(u) + ';'
  end;

UTF8toHTML := e
end;

function UTF8toLatin1(const s: string): mystring;
var
  f, e : mystring;
  u: Unicode;
begin
f := s;
e := '';

while f <> '' do
  begin
  u := getUTF8char(f);
  if u<=255 
    then begin
         if (not ISO8859ControlChar(u)) or (u=ord(unknownChar))
            then e := e + chr(u)
         end
    else e := e + unknownChar
  end;

UTF8toLatin1 := e
end;

function UTF8toLatin9(const s: string): mystring;
var
  f, e : mystring;
  u: Unicode;
begin
f := s;
e := '';

while f <> '' do
  begin
  u := getUTF8char(f);
  if u<=255 
    then begin
         if (not ISO8859ControlChar(u)) or (u=ord(unknownChar))
            then e := e + chr(u) { simply ignore replaced chars }
         end
    else 
      case u of
        $20AC : e := e + chr($A4); { euro }
        $0160 : e := e + chr($A6); { S with caron }
        $0161 : e := e + chr($A8); { s with caron }
        $017D : e := e + chr($B4); { Z with caron }
        $017E : e := e + chr($B8); { z with caron }
        $0152 : e := e + chr($BC); { ligature OE }
        $0153 : e := e + chr($BD); { ligature oe }
        $0178 : e := e + chr($BE); { Y with diaeresis }
        otherwise e := e + unknownChar
        end
  end;

UTF8toLatin9 := e
end;

function Latin1toHTML(const s: string): mystring;
var
  i : integer;
  e : mystring;
begin
{ HTML entities are coded as unicode.
  But Latin1 code-numbers are equivalent with unicode (subset) }
e := '';
for i := 1 to length(s) do
  if s[i] < #$80
     then e := e + s[i]
     else e := e + '&#' + IntToStr(ord(s[i])) + ';';

Latin1toHTML := e
end;

function Latin1toOEM(const s: string): mystring;
var
  i : integer;
  e : mystring;
  c : char;
begin
e := s;
for i := 1 to length(e) do
  if e[i] >= #$80 then
    for c:=#$80 to #$FF do
      if CP850Unicode[c]=ord(e[i]) then e[i] := c;
Latin1toOEM := e
end;


function OEMtoLatin1(const s: string): mystring;
var
  i : integer;
  e : mystring;
begin
e := s;
for i := 1 to length(e) do
  if e[i] >= #$80 then
    if CP850Unicode[e[i]]<=$00FF
      then e[i] := chr(CP850Unicode[e[i]])
      else e[i] := unknownChar;
OEMtoLatin1 := e
end;

function CP1252toLatin9(const s: string): mystring;
var
  i: integer;
  e: mystring;
begin
e := s;
for i := 1 to length(e) do
  case ord(e[i]) of
    $80 : e[i] := chr($A4);
    $8A : e[i] := chr($A6);
    $9A : e[i] := chr($A8);
    $8E : e[i] := chr($B4);
    $9E : e[i] := chr($B8);
    $8C : e[i] := chr($BC);
    $9C : e[i] := chr($BD);
    $9F : e[i] := chr($BE);
    { similar looking chars: }
    $A6 : e[i] := '|';
    $A8 : e[i] := '"';
    $B4 : e[i] := '''';
    $B8 : e[i] := ',';
    { unsupported: }
    $A4, $BC, $BD, $BE : e[i] := unknownChar;
    end;
CP1252toLatin9 := e
end;

function OEMtoUTF8(const s: string): mystring;
var
  i : integer;
  e : mystring;
begin
e := '';
for i := 1 to length(s) do
  if s[i]<=#$80
    then e := e + s[i]
    else e := e + EncodeUTF8(CP850Unicode[s[i]]);
OEMtoUTF8 := e
end;

function UTF8toOEM(const s: string): mystring;
var
  f, e : mystring;
  c: char;
  u: Unicode;
begin
f := s;
e := '';

while f <> '' do
  begin
  u := getUTF8char(f);
  if u<=$0080
    then e := e + chr(u)
    else for c:=#$80 to #$FF do
           if CP850Unicode[c]=u then e := e + c
  end;

UTF8toOEM := e
end;

function UTF8Length(const s: string): LongInt;
var i, res: LongInt;
begin
res := 0;
{ count ASCII bytes and start bytes, ignore the rest }
for i := 1 to length(s) do
   if (ord(s[i])<=127) or (ord(s[i])>=$C0) then inc(res);
UTF8Length := res
end;

function basename(const s: string): mystring;
begin
{$IfDef __GPC__}
  basename := NameExtFromPath(s)
{$Else}
  basename := ExtractFileName(s)
{$EndIf}
end;

{$IfDef __GPC__}

  function gethtmlname(s: mystring): mystring;
  var path, name, ext: mystring;
  begin
  FSplit(s, path, name, ext);
  gethtmlname := name + '.html'
  end;

{$Else}

  function gethtmlname(s: mystring): mystring;
  begin
  gethtmlname := ChangeFileExt(ExtractFileName(s), '.html')
  end;

{$EndIf}

{ gets string after colon, leading spaces stripped }
function getvalue(x: string): mystring;
var i: integer;
begin
i := succ(pos(':', x));
while (x[i]=' ') or (x[i]=TAB) do inc(i);
getvalue := copy(x, i, length(x)-i+1);
end;

function getQuizTitle(const x: string): MyString;
var 
  inp: text;
  s, e : String80;
  title,
  charset: String80;
  ignore : integer;
{$IfDef FPC}
  var Buffer : array[1..1024] of char;
{$EndIf}
begin
title := '';
charset := '';
Assign(inp, x);
{$IfDef FPC}
  SetTextBuf(inp, Buffer);
{$EndIf}
reset(inp);
repeat
  ReadLn(inp, s);
  s := stripWhitespace(s);
  e := makeUpcase(s);
  if (pos('TITLE:', e)<>0) or
     (pos('TITEL:', e)<>0) then title := getvalue(s);
  if (pos('CHARSET:',e)=1) or
     (pos('ZEICHENSATZ:',e)=1) then charset := getvalue(s)
until ((title<>'') and (charset<>'')) or EOF(inp) or (IOResult<>0);
close(inp);
ignore := IOResult; { ignore errors }

if title<>'' 
  then 
    begin
    if charset='' 
      then charset:='US-ASCII'
      else charset := makeUpcase(charset);             
    if checkASCII(charset)  then title := forceASCII(title);
    if checkISO(charset)    then title := Latin1ToUTF8(title);
    if checkCP1252(charset) then title := Latin1ToUTF8(title); {@@@}
    if checkOEM(charset)    then title := OEMtoUTF8(title); 
    getQuizTitle := title 
    end
  else getQuizTitle := x { no title -> filename }
end;

{$IfDef __GPC__}
  function ListEntries(const dir, ext: string; showentry: Tshowentry): boolean;
  var
    d: DirPtr;
    s: TString;
    found: boolean;
  begin
  found := false;
  if dir=''
    then d := OpenDir('.')
    else d := OpenDir(dir);
  if IOResult=0 then
    begin
    s := ReadDir(d);
    while s<>'' do
      begin
      if (s[length(s)-length(ext)+1 .. length(s)] = ext)
        {$IfDef DJGPP}
	  or ((length(ext)>4) and 
	      (s[length(s)-3 .. length(s)] = ext[1..4]))
	{$EndIf}
        then
         begin
         showentry(dir, s);
         found := true
         end;
      s := ReadDir(d)
      end
    end;
  CloseDir(d);
  ListEntries := found
  end;

{$Else} { not GPC }

  function ListEntries(const dir, ext: string; showentry: Tshowentry): boolean;
  var
    info  : TSearchRec;
    rslt  : LongInt;
    found : boolean;
  begin
  found := false;

  if dir=''
     then rslt := SysUtils.FindFirst('*'+ext, FaAnyFile, info)
     else rslt := SysUtils.FindFirst(dir+DirectorySeparator+'*'+ext, FaAnyFile, info);
  if rslt=0 then found := true;

  while rslt=0 do
     begin
     showentry(dir, info.name);
     rslt := SysUtils.FindNext(info)
     end;
  SysUtils.FindClose(info);

  ListEntries := found
  end;
{$EndIf} { not __GPC__}

procedure freeQuizfileList;
var previous: PquizfileList;
begin
while quizfileList<>NIL do
  begin
  previous := quizfileList;
  quizfileList := quizfileList^.next;
  dispose(previous)
  end;
quizFileListEnd := NIL
end;

procedure addQuizFileListEntry(const dir, f : string);
var previous : PquizfileList;
begin
previous := quizfileListEnd;
new(quizfileListEnd);

if quizfileList=NIL 
  then quizfileList := quizfileListEnd
  else previous^.next := quizfileListEnd;

with quizfileListEnd^ do
  begin
  if dir='' 
    then filename := f
    else filename := dir+DirectorySeparator+f;
  title := getQuizTitle(quizfileListEnd^.filename);
  next := NIL
  end
end;

procedure addQuizFileList(const dir, ext: string);
var ignore: boolean;
begin
ignore := ListEntries(dir, ext, addQuizFileListEntry)
end;

procedure fillQuizfileList; { uses quizpath! }
var dir, path: mystring;
begin
if quizfileList<>NIL then freeQuizfileList;

path := getquizpath;
while path<>'' do
  begin
  dir := getnextdir(path);
  addQuizfileList(dir, quizext);
  addQuizfileList(dir, quizext2);
  end
end;

function searchQuizfileList(searchNr: integer): mystring;
var
  entry : PquizfileList;
  Nr: integer;
begin
Nr := 1;
entry := quizfileList;
while (Nr<>searchNr) and (entry<>NIL) do
  begin
  entry := entry^.next;
  inc(Nr)
  end;

{ hopefully found }
if entry<>NIL 
  then searchQuizfileList := entry^.filename
  else searchQuizfileList := '';
end;

procedure nobreak;
begin
{ SetCBreak(false) } {@@@ too DOS specific }
end;

procedure setLFNsupport;
begin
{$IfDef FPC}
  {$IfDef Go32v2}
    LFNsupport := true;
    FileNameCaseSensitive := true
  {$EndIf}
{$EndIf}
end;

function checkISO(const s: mystring): boolean;
begin
checkISO := (pos('ISO-8859',s)<>0) or
            (pos('ISO 8859',s)<>0) or 
            (pos('ISO_8859',s)<>0) or
            (pos('LATIN',s)<>0) 
end;

function checkCP1252(const s: mystring): boolean;
begin
checkCP1252 := (s='WINDOWS-1252') or (s='CP1252')
end;

function checkUTF8(const s: mystring): boolean;
begin
checkUTF8 := (pos('UTF-8',s)<>0) or 
             (pos('UTF8',s)<>0)
end;

function checkOEM(const s: mystring): boolean;
begin
checkOEM := (pos('DOS',s)<>0) or { 'DOS' is not official! }
            (pos('CP850',s)<>0) or 
            (pos('CP437',s)<>0) or
            (pos('IBM850',s)<>0) or 
            (pos('IBM437',s)<>0) or
            (s='850') or 
            (s='437')
end;

function checkASCII(const s: mystring): boolean;
begin
checkASCII := (s='ASCII') or 
              (s='US-ASCII')
end;

function checkUTF8display: boolean;
var
 UTF8 : boolean;
 s: mystring;
begin
UTF8 := false;

s := makeUpcase(GetEnvironmentVariable('MM_CHARSET'));
if (s='UTF-8') or (s='UTF8') then UTF8 := true;

s := makeUpcase(GetEnvironmentVariable('LANG'));
if (pos('UTF-8', s)<>0) or (pos('UTF8', s)<>0) then UTF8 := true;

checkUTF8display := UTF8
end;

function checkOEMdisplay: boolean;
var s: mystring;
begin
{ Windows uses different charsets in the Editor and in console mode }
{$IfDef Win32} {$Define OEM} {$EndIf}
{$IfDef DPMI} {$Define OEM} {$EndIf}
{$IfDef __OS_DOS__} {$Define OEM} {$EndIf}
{$IfDef MSDOS}      {$Define OEM} {$EndIf}

{$IfDef OEM}
  checkOEMdisplay := true;
{$Else}
  s := makeUpcase(GetEnvironmentVariable('MM_CHARSET'));
  checkOEMdisplay := (s='DOS') { 'DOS' is not official! }
                     or (s='CP850') or (s='CP437')
                     or (s='IBM850')or (s='IBM437')
                     or (s='850') or (s='437')
{$EndIf}
end;

function IntTo2Str(i: LongInt): mystring;
begin
if i<10 
  then IntTo2Str := '0'+IntToStr(i)
  else IntTo2Str := IntToStr(i)
end;


function ShowTime(sec: LongInt): mystring;
var h, m, s: integer;
begin
h := sec div (60*60);
m := (sec div 60) mod 60;
s := sec mod 60;
if h=0 
  then ShowTime := IntTo2Str(m)+':'+IntTo2Str(s)
  else ShowTime := IntToStr(h)+':'+IntTo2Str(m)+':'+IntTo2Str(s)
end;

function GetSecs: LongInt;
begin
{$IfDef FPC}
  GetSecs := trunc(TimestampToMSecs(DateTimeToTimestamp(Now))) div 1000
{$Else}
  GetSecs := GetMicroSecondTime div 1000000
{$EndIf}
end;

{$IfDef __GPC__}
  function IntToStr(i: LongInt): mystring;
  begin
  IntToStr := Integer2String(i)
  end;
{$EndIf}



INITIALIZATION

  unknownChar := '?';
  disableSignals; { initializes Signals }
  
  quizfileList := NIL;
  quizfileListEnd := NIL;

  QUIZPATH := getParamDir; { highest priority }
  if QUIZPATH='' then QUIZPATH := GetEnvironmentVariable('QUIZPATH');
  if QUIZPATH='' then 
    QUIZPATH := getprefix+DirectorySeparator+'share'+DirectorySeparator+
                'akfquiz'+PathSeparator+'.';

  {$IfDef Go32v2}
    { Compiler checks LFNsupport just on drive C: - that's stupid! :-( }
    if GetEnvironmentVariable('LFN')<>'' then setLFNsupport;

    {$IfDef ForceLFN}
      setLFNsupport;
      {$Info LFN support enforced}
    {$EndIf}
  {$EndIf}


FINALIZATION

  freeQuizFileList

end.
